FIXTURE_UTILIZATION_WARNING = (
    "Warning: Utilization attributes configuration has no effect until cluster "
    "property option 'placement-strategy' is set to one of the values: "
    "'balanced', 'minimal', 'utilization'\n"
)
